# type: ignore
import pyodide
import js
from typing import Optional, List, Dict
from resotoclient import rnd_str
from resotoclient.jwt_utils import encode_jwt_to_headers
from resotonotebook.client import Client
from resotoclient.models import JsObject, JsValue


class PyodideResotoClient(Client):
    def __init__(self, url: str, psk: Optional[str]):
        self.url = url
        self.psk = psk
        self.session_id = rnd_str()

    def headers(self) -> Dict[str, str]:

        headers = {"Content-type": "application/json", "Accept": "application/json"}

        if self.psk:
            encode_jwt_to_headers(headers, {}, self.psk)

        return headers

    async def search_list(self, search: str, section: Optional[str], graph: str) -> List[JsObject]:
        path = f"/graph/{graph}/search/list"

        url = js.URL.new(path, js.location.origin)
        if section:
            url.searchParams.append("section", section)
        url.searchParams.append("session_id", self.session_id)

        headers = self.headers()

        options = {"method": "POST", "body": search, "headers": headers}

        async def call(url, options):
            response = await js.fetch(url, options)

            if response.status == 200:
                json = await response.json()
                return json.to_py()
            else:
                text = await response.text()
                raise AttributeError(text)

        options = pyodide.ffi.to_js(options, dict_converter=js.Object.fromEntries)

        return await call(url, options)

    async def search_graph(self, search: str, section: Optional[str], graph: str) -> List[JsObject]:
        path = f"/graph/{graph}/search/graph"

        url = js.URL.new(path, js.location.origin)
        if section:
            url.searchParams.append("section", section)
        url.searchParams.append("session_id", self.session_id)

        headers = self.headers()

        options = {"method": "POST", "body": search, "headers": headers}

        async def call(url, options):
            response = await js.fetch(url, options)

            if response.status == 200:
                json = await response.json()
                return json.to_py()
            else:
                text = await response.text()
                raise AttributeError(text)

        options = pyodide.ffi.to_js(options, dict_converter=js.Object.fromEntries)

        return await call(url, options)

    async def cli_execute(self, query: str, graph: str, section: Optional[str]) -> List[JsValue]:
        path = "/cli/execute"

        url = js.URL.new(path, js.location.origin)
        if section:
            url.searchParams.append("section", section)
        url.searchParams.append("graph", graph)
        url.searchParams.append("session_id", self.session_id)

        headers = self.headers()
        headers["Content-type"] = "text/plain"

        options = {"method": "POST", "body": query, "headers": headers}

        async def call(url, options):
            response = await js.fetch(url, options)

            if response.status == 200:
                content_type = response.headers.get("Content-Type")
                if content_type == "text/plain":
                    text = await response.text()
                    return [text.to_py()]
                elif content_type == "application/json":
                    json = (await response.json()).to_py()
                    if isinstance(json, list):
                        return json
                    return [json]
                else:
                    msg = f"Content type {content_type} is not supported in "
                    +"jupyterlite version of resotonotebook. Use a jupyter notebook instead"
                    raise NotImplementedError(msg)
            else:
                raise AttributeError(response.text)

        options = pyodide.ffi.to_js(options, dict_converter=js.Object.fromEntries)

        return await call(url, options)



from typing import Dict, Optional, List
import pandas as pd
from resotoclient.models import JsObject, JsValue
import graphviz
from enum import Enum
from dataclasses import dataclass
from collections import defaultdict
from resotonotebook.client import get_client
import sys


class ResotoNotebook:
    def __init__(self, url: str, psk: Optional[str], resoto_graph: str = "resoto") -> None:
        self.client = get_client(url, psk)
        self.resoto_graph = resoto_graph

    @staticmethod
    def with_psk(psk: Optional[str]) -> "ResotoNotebook":
        return ResotoNotebook("http://localhost:8900", psk)

    async def search(self, query: str, section: Optional[str] = "reported") -> pd.DataFrame:
        iter = await self.client.search_list(search=query, section=section, graph=self.resoto_graph)

        def extract_node(node: JsObject) -> Optional[JsObject]:
            reported = node.get("reported")
            if not isinstance(reported, Dict):
                return None
            reported["account_id"] = js_find(
                node,
                ["ancestors", "account", "reported", "id"],
            )
            reported["region_id"] = js_find(
                node,
                ["ancestors", "region", "reported", "id"],
            )
            reported["cloud_id"] = js_find(
                node,
                ["ancestors", "cloud", "reported", "id"],
            )
            return reported

        nodes = [extract_node(node) for node in iter]
        return pd.DataFrame(nodes)

    async def graph(
        self,
        query: str,
        section: Optional[str] = "reported",
        engine: str = "sfdp",
        format: str = "svg",
    ) -> graphviz.Digraph:
        digraph = graphviz.Digraph(comment=query)
        digraph.format = format
        digraph.engine = engine
        digraph.graph_attr = {"rankdir": "LR", "splines": "true", "overlap": "false"}  # type: ignore
        digraph.node_attr = {  # type: ignore
            "shape": "plain",
            "colorscheme": "paired12",
        }
        cit = iter(range(0, sys.maxsize))
        colors: Dict[str, int] = defaultdict(lambda: (next(cit) % 12) + 1)

        results = await self.client.search_graph(search=query, section=section, graph=self.resoto_graph)

        for elem in results:
            if elem.get("type") == "node":
                kind = js_get(elem, ["reported", "kind"])
                color = colors[kind]
                rd = ResourceDescription(
                    id=js_get(elem, ["reported", "id"]),
                    name=js_get(elem, ["reported", "name"]),
                    uid=js_get(elem, ["id"]),
                    kind=parse_kind(kind),
                    kind_name=kind,
                )
                digraph.node(  # type: ignore
                    name=js_get(elem, ["id"]),
                    # label=rd.name,
                    label=render_resource(rd, color),
                    shape="plain",
                )
            elif elem.get("type") == "edge":
                digraph.edge(js_get(elem, ["from"]), js_get(elem, ["to"]))  # type: ignore

        return digraph

    async def cli_execute(self, query: str, section: str = "reported") -> pd.DataFrame:
        iter = await self.client.cli_execute(query, self.resoto_graph, section)

        normalize = False

        def extract_node(node: JsValue) -> Optional[JsValue]:
            if isinstance(node, dict):
                reported = node.get("reported")
                if not isinstance(reported, Dict):
                    nonlocal normalize
                    normalize = True
                    return node
                reported["account_id"] = js_find(
                    node,
                    ["ancestors", "account", "reported", "id"],
                )
                reported["region_id"] = js_find(
                    node,
                    ["ancestors", "region", "reported", "id"],
                )
                reported["cloud_id"] = js_find(
                    node,
                    ["ancestors", "cloud", "reported", "id"],
                )
                return reported
            return node

        nodes = [extract_node(node) for node in iter]
        if normalize:
            return pd.json_normalize(nodes)  # type: ignore
        else:
            return pd.DataFrame(nodes)


class ResourceKind(Enum):
    UNKNOWN = 1
    INSTANCE = 2
    VOLUME = 3
    IMAGE = 4
    FIREWALL = 5
    K8S_CLUSER = 6
    NETWORK = 7
    LOAD_BALANCER = 8
    CLOUD = 9


kind_colors = {
    ResourceKind.INSTANCE: "8",
    ResourceKind.VOLUME: "4",
    ResourceKind.IMAGE: "7",
    ResourceKind.FIREWALL: "6",
    ResourceKind.K8S_CLUSER: "5",
    ResourceKind.NETWORK: "10",
    ResourceKind.LOAD_BALANCER: "9",
    ResourceKind.CLOUD: "1",
}


@dataclass
class ResourceDescription:
    uid: str
    name: str
    id: str
    kind: ResourceKind
    kind_name: str


def render_resource(
    resource: ResourceDescription,
    color: int,
) -> str:
    return f"""\
<<TABLE STYLE="ROUNDED" COLOR="{color}" BORDER="3" CELLBORDER="1" CELLPADDING="5">
    <TR>
        <TD SIDES="B">{resource.kind_name}</TD>
    </TR>
    <TR>
        <TD SIDES="B">{resource.id}</TD>
    </TR>
    <TR>
        <TD BORDER="0">{resource.name}</TD>
    </TR>
</TABLE>>"""


do_kinds = {
    "droplet": ResourceKind.INSTANCE,
    "volume": ResourceKind.VOLUME,
    "image": ResourceKind.IMAGE,
    "firewall": ResourceKind.FIREWALL,
    "kubernetes_cluster": ResourceKind.K8S_CLUSER,
    "network": ResourceKind.NETWORK,
    "load_balancer": ResourceKind.LOAD_BALANCER,
}


def parse_kind(kind: str) -> ResourceKind:
    cloud, rest = kind.split("_")[0], "_".join(kind.split("_")[1:])
    if cloud == "digitalocean":
        return do_kinds.get(rest) or ResourceKind.UNKNOWN
    else:
        return ResourceKind.UNKNOWN


def js_find(node: JsObject, path: List[str]) -> Optional[str]:
    """
    Get a value in a nested dict.
    """
    if len(path) == 0:
        return None
    else:
        value = node.get(path[0])
        if len(path) == 1:
            return value if isinstance(value, str) else None
        if not isinstance(value, dict):
            return None
        return js_find(value, path[1:])


def js_get(node: JsObject, path: List[str]) -> str:
    result = js_find(node, path)
    if result is None:
        raise ValueError(f"Path {path} not found in {node}")
    return result



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['resotonotebook']

package_data = \
{'': ['*']}

install_requires = \
['graphviz>=0.20,<0.21',
 'nbformat>=5.3.0,<6.0.0',
 'pandas-stubs>=1.2.0,<2.0.0',
 'pandas>=1.4.2,<2.0.0',
 'plotly>=5.7.0,<6.0.0',
 'resotoclient>=0.1.6,<0.2.0']

setup_kwargs = {
    'name': 'resotonotebook',
    'version': '0.2.2',
    'description': 'Resoto Python client library',
    'long_description': '# resotonotebook\nSmall library for using Resoto with Jupyter Notebooks.\n\n## Installation\n\n```bash\npip install resotonotebook\n```\n\n## Usage\n\n```python\nfrom resotonotebook import ResotoNotebook\nrnb = ResotoNotebook("https://localhost:8900", None)\nrnb.search("is(instance)").groupby(["kind"])["kind"].count()\n```\n```\nkind\naws_ec2_instance        497\ndigitalocean_droplet      7\nexample_instance          2\ngcp_instance             12\nName: kind, dtype: int64\n```\n\nFor more see the notebook in the examples directory.\n',
    'author': 'Some Engineering Inc.',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/someengineering/resotonotebook',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from abc import ABC, abstractmethod

from typing import List, Optional
from resotoclient import ResotoClient
from resotoclient.models import JsObject, JsValue
import sys


class Client(ABC):
    @abstractmethod
    async def search_list(self, search: str, section: Optional[str], graph: str) -> List[JsObject]:
        pass

    @abstractmethod
    async def search_graph(self, search: str, section: Optional[str], graph: str) -> List[JsObject]:
        pass

    @abstractmethod
    async def cli_execute(self, query: str, graph: str, section: Optional[str]) -> List[JsValue]:
        pass


class AsyncResotoClient(Client):
    def __init__(self, client: ResotoClient):
        self.client = client

    async def search_list(self, search: str, section: Optional[str], graph: str) -> List[JsObject]:
        return list(self.client.search_list(search=search, section=section, graph=graph))

    async def search_graph(self, search: str, section: Optional[str], graph: str) -> List[JsObject]:
        return list(self.client.search_graph(search=search, section=section, graph=graph))

    async def cli_execute(self, query: str, graph: str, section: Optional[str]) -> List[JsValue]:
        return list(self.client.cli_execute(command=query, graph=graph, section=section))


def get_client(url: str, psk: Optional[str]) -> Client:
    if sys.platform == "emscripten":
        # import on this level to avoid loading pyodide in the cpython context
        from resotonotebook.pyodide_client import PyodideResotoClient

        return PyodideResotoClient(url, psk)
    else:
        return AsyncResotoClient(ResotoClient(url, psk))
